@@ -8,7 +8,7 @@ const char * dummy[] = {
 "/** @file version.h",
 " * @brief Define preprocessor symbols for the library version",
 " */",
-"// Copyright (C) 2002,2004,2005,2006,2007,2008,2009,2010,2011,2012,2013,2015,2016,2017 Olly Betts",
+"// Copyright (C) 2002,2004,2005,2006,2007,2008,2009,2010,2011,2012,2013,2015,2016,2017,2018 Olly Betts",
 "//",
 "// This program is free software; you can redistribute it and/or",
 "// modify it under the terms of the GNU General Public License as",
@@ -215,5 +215,22 @@ const char * dummy[] = {
 "      (XAPIAN_MINOR_VERSION > (B) || \\",
 "       (XAPIAN_MINOR_VERSION == (B) && XAPIAN_REVISION >= (C)))))",
 "",
+"/// We support move semantics when we're confident the compiler supports it.",
+"///",
+"/// C++11 move semantics are very useful in threaded code that wants to",
+"/// hand-off Xapian objects to worker threads, but in this case it's very",
+"/// unhelpful for availability of these semantics to vary by compiler as it",
+"/// quietly leads to a build with non-threadsafe behaviour.",
+"///",
+"/// User code can #define XAPIAN_MOVE_SEMANTICS to force this on, and will",
+"/// then get a compilation failure if the compiler lacks suitable support.",
+"#ifndef XAPIAN_MOVE_SEMANTICS",
+"# if __cplusplus >= 201103L || \\",
+"     (defined _MSC_VER && _MSC_VER >= 1900) || \\",
+"     defined XAPIAN_LIB_BUILD",
+"#  define XAPIAN_MOVE_SEMANTICS",
+"# endif",
+"#endif",
+"",
 "#endif /* XAPIAN_INCLUDED_VERSION_H */"
 };
