@@ -9,7 +9,7 @@ typedef enum {
     e_unexpected_token = 1,
     e_string_omitted = 2,
     e_unexpected_token_in_among = 3,
-    // For codes above here, report "after " t->previous_token after the error.
+    /* For codes above here, report "after " t->previous_token after the error. */
     e_unresolved_substring = 14,
     e_not_allowed_inside_reverse = 15,
     e_empty_grouping = 16,
@@ -17,7 +17,7 @@ typedef enum {
     e_empty_among = 18,
     e_adjacent_bracketed_in_among = 19,
     e_substring_preceded_by_substring = 20,
-    // For codes below here, tokeniser->b is printed before the error.
+    /* For codes below here, tokeniser->b is printed before the error. */
     e_redeclared = 30,
     e_undeclared = 31,
     e_declared_as_different_mode = 32,
@@ -35,8 +35,6 @@ static struct node * read_C(struct analyser * a);
 static struct node * C_style(struct analyser * a, const char * s, int token);
 
 
-static void fault(int n) { fprintf(stderr, "fault %d\n", n); exit(1); }
-
 static void print_node_(struct node * p, int n, const char * s) {
 
     int i;
@@ -76,22 +74,37 @@ static struct node * new_node(struct analyser * a, int type) {
 
 static const char * name_of_mode(int n) {
     switch (n) {
-         default: fault(0);
-         case m_backward: return "string backward";
-         case m_forward:  return "string forward";
-    /*   case m_integer:  return "integer";  */
+        case m_backward: return "string backward";
+        case m_forward:  return "string forward";
+    /*  case m_integer:  return "integer";  */
     }
+    fprintf(stderr, "Invalid mode %d in name_of_mode()\n", n);
+    exit(1);
 }
 
 static const char * name_of_type(int n) {
     switch (n) {
-         default: fault(1);
-         case 's': return "string";
-         case 'i': return "integer";
-         case 'r': return "routine";
-         case 'R': return "routine or grouping";
-         case 'g': return "grouping";
+        case 's': return "string";
+        case 'i': return "integer";
+        case 'r': return "routine";
+        case 'R': return "routine or grouping";
+        case 'g': return "grouping";
     }
+    fprintf(stderr, "Invalid type %d in name_of_type()\n", n);
+    exit(1);
+}
+
+static const char * name_of_name_type(int code) {
+    switch (code) {
+        case t_string: return "string";
+        case t_boolean: return "boolean";
+        case t_integer: return "integer";
+        case t_routine: return "routine";
+        case t_external: return "external";
+        case t_grouping: return "grouping";
+    }
+    fprintf(stderr, "Invalid type code %d in name_of_name_type()\n", code);
+    exit(1);
 }
 
 static void count_error(struct analyser * a) {
@@ -110,6 +123,7 @@ static void error2(struct analyser * a, error_code n, int x) {
             fprintf(stderr, "%s omitted", name_of_token(t->omission)); break;
         case e_unexpected_token_in_among:
             fprintf(stderr, "in among(...), ");
+            /* fall through */
         case e_unexpected_token:
             fprintf(stderr, "unexpected %s", name_of_token(t->token));
             if (t->token == c_number) fprintf(stderr, " %d", t->number);
@@ -228,13 +242,24 @@ static void check_routine_mode(struct analyser * a, struct name * p, int mode) {
 
 static void check_name_type(struct analyser * a, struct name * p, int type) {
     switch (type) {
-        case 's': if (p->type == t_string) return; break;
-        case 'i': if (p->type == t_integer) return; break;
-        case 'b': if (p->type == t_boolean) return; break;
-        case 'R': if (p->type == t_grouping) return;
-        case 'r': if (p->type == t_routine ||
-                      p->type == t_external) return; break;
-        case 'g': if (p->type == t_grouping) return; break;
+        case 's':
+            if (p->type == t_string) return;
+            break;
+        case 'i':
+            if (p->type == t_integer) return;
+            break;
+        case 'b':
+            if (p->type == t_boolean) return;
+            break;
+        case 'R':
+            if (p->type == t_grouping) return;
+            /* FALLTHRU */
+        case 'r':
+            if (p->type == t_routine || p->type == t_external) return;
+            break;
+        case 'g':
+            if (p->type == t_grouping) return;
+            break;
     }
     error2(a, e_not_of_type_x, type);
 }
@@ -278,10 +303,13 @@ handle_as_name:
                     p->referenced = false;
                     p->used_in_among = false;
                     p->used = 0;
+                    p->value_used = false;
+                    p->initialised = false;
                     p->local_to = 0;
                     p->grouping = 0;
                     p->definition = 0;
-                    a->name_count[type] ++;
+                    p->declaration_line_number = t->line_number;
+                    a->name_count[type]++;
                     p->next = a->names;
                     a->names = p;
                     if (token != c_name) {
@@ -358,8 +386,18 @@ static struct node * read_AE(struct analyser * a, int B) {
     struct node * q;
     switch (read_token(t)) {
         case c_minus: /* monadic */
+            q = read_AE(a, 100);
+            if (q->type == c_neg) {
+                /* Optimise away double negation, which avoids generators
+                 * having to worry about generating "--" (decrement operator
+                 * in many languages).
+                 */
+                p = q->right;
+                /* Don't free q, it's in the linked list a->nodes. */
+                break;
+            }
             p = new_node(a, c_neg);
-            p->right = read_AE(a, 100);
+            p->right = q;
             break;
         case c_bra:
             p = read_AE(a, 0);
@@ -368,10 +406,12 @@ static struct node * read_AE(struct analyser * a, int B) {
         case c_name:
             p = new_node(a, c_name);
             name_to_node(a, p, 'i');
+            if (p->name) p->name->value_used = true;
             break;
         case c_maxint:
         case c_minint:
             a->int_limits_used = true;
+            /* fall through */
         case c_cursor:
         case c_limit:
         case c_len:
@@ -617,6 +657,8 @@ static struct node * read_among(struct analyser * a) {
                 q = read_C_list(a); break;
             default:
                 error(a, e_unexpected_token_in_among);
+                previous_token = token;
+                continue;
             case c_ket:
                 if (p->number == 0) error(a, e_empty_among);
                 if (t->error_count == 0) make_among(a, p, substring);
@@ -680,14 +722,18 @@ static struct node * read_C(struct analyser * a) {
         case c_loop:
         case c_atleast:
             return C_style(a, "AC", token);
-        case c_setmark:
-            return C_style(a, "i", token);
+        case c_setmark: {
+            struct node * n = C_style(a, "i", token);
+            if (n->name) n->name->initialised = true;
+            return n;
+        }
         case c_tomark:
         case c_atmark:
         case c_hop:
             return C_style(a, "A", token);
         case c_delete:
             check_modifyable(a);
+            /* fall through */
         case c_next:
         case c_tolimit:
         case c_atlimit:
@@ -698,20 +744,31 @@ static struct node * read_C(struct analyser * a) {
         case c_debug:
             return C_style(a, "", token);
         case c_assignto:
-        case c_sliceto:
+        case c_sliceto: {
+            struct node *n;
             check_modifyable(a);
-            return C_style(a, "s", token);
+            n = C_style(a, "s", token);
+            if (n->name) n->name->initialised = true;
+            return n;
+        }
         case c_assign:
         case c_insert:
         case c_attach:
-        case c_slicefrom:
+        case c_slicefrom: {
+            struct node *n;
             check_modifyable(a);
-            return C_style(a, "S", token);
+            n = C_style(a, "S", token);
+            if (n->name) n->name->value_used = true;
+            return n;
+        }
         case c_setlimit:
             return C_style(a, "CfD", token);
         case c_set:
-        case c_unset:
-            return C_style(a, "b", token);
+        case c_unset: {
+            struct node * n = C_style(a, "b", token);
+            if (n->name) n->name->initialised = true;
+            return n;
+        }
         case c_dollar:
             get_token(a, c_name);
             {
@@ -722,8 +779,18 @@ static struct node * read_C(struct analyser * a) {
                 switch (q ? q->type : t_string)
                     /* above line was: switch (q->type) - bug #1 fix 7/2/2003 */
                 {
-                    default: error(a, e_not_of_type_string_or_integer);
+                    default:
+                        error(a, e_not_of_type_string_or_integer);
+                        /* Handle $foo for unknown 'foo' as string since
+                         * that's more common and so less likely to cause
+                         * an error avalanche. */
+                        /* fall through */
                     case t_string:
+			/* Assume for now that $ on string both initialises and
+			 * uses the string variable.  FIXME: Can we do better?
+			 */
+			q->initialised = true;
+			q->value_used = true;
                         a->mode = m_forward;
                         a->modifyable = true;
                         p = new_node(a, c_dollar);
@@ -731,7 +798,27 @@ static struct node * read_C(struct analyser * a) {
                     case t_integer:
                     /*  a->mode = m_integer;  */
                         p = new_node(a, read_AE_test(a));
-                        p->AE = read_AE(a, 0); break;
+                        p->AE = read_AE(a, 0);
+                        if (q) {
+                            /* +=, etc don't "initialise" as they only amend an
+                             * existing value.  Similarly, they don't count as
+                             * using the value.
+                             */
+                            switch (p->type) {
+                                case c_mathassign:
+                                    q->initialised = true;
+                                    break;
+                                case c_eq:
+                                case c_ne:
+                                case c_gr:
+                                case c_ge:
+                                case c_ls:
+                                case c_le:
+                                    q->value_used = true;
+                                    break;
+                            }
+                        }
+                        break;
                 }
                 if (q) mark_used_in(a, q, p);
                 p->name = q;
@@ -747,10 +834,14 @@ static struct node * read_C(struct analyser * a) {
                     mark_used_in(a, q, p);
                     switch (q->type) {
                         case t_boolean:
-                            p->type = c_booltest; break;
+                            p->type = c_booltest;
+                            q->value_used = true;
+                            break;
                         case t_integer:
                             error(a, e_misplaced); /* integer name misplaced */
+                            break;
                         case t_string:
+                            q->value_used = true;
                             break;
                         case t_routine:
                         case t_external:
@@ -824,10 +915,11 @@ static void read_define_grouping(struct analyser * a, struct name * q) {
         NEW(grouping, p);
         if (a->groupings == 0) a->groupings = p; else a->groupings_end->next = p;
         a->groupings_end = p;
-        q->grouping = p;
+        if (q) q->grouping = p;
         p->next = 0;
         p->name = q;
-        p->number = q->count;
+        p->number = q ? q->count : 0;
+        p->line_number = a->tokeniser->line_number;
         p->b = create_b(0);
         while (true) {
             switch (read_token(t)) {
@@ -841,7 +933,7 @@ static void read_define_grouping(struct analyser * a, struct name * q) {
                     }
                     break;
                 case c_literalstring:
-                    p->b = alter_grouping(p->b, t->b, style, a->utf8);
+                    p->b = alter_grouping(p->b, t->b, style, (a->encoding == ENC_UTF8));
                     break;
                 default: error(a, e_unexpected_token); return;
             }
@@ -885,8 +977,8 @@ static void read_define_routine(struct analyser * a, struct name * q) {
     if (q) q->definition = p->left;
 
     if (a->substring != 0) {
-         error2(a, e_unresolved_substring, a->substring->line_number);
-         a->substring = 0;
+        error2(a, e_unresolved_substring, a->substring->line_number);
+        a->substring = 0;
     }
     p->amongvar_needed = a->amongvar_needed;
 }
@@ -894,8 +986,26 @@ static void read_define_routine(struct analyser * a, struct name * q) {
 static void read_define(struct analyser * a) {
     if (get_token(a, c_name)) {
         struct name * q = find_name(a);
-        if (q != 0 && q->type == t_grouping) read_define_grouping(a, q);
-            else read_define_routine(a, q);
+        int type;
+        if (q) {
+            type = q->type;
+        } else {
+            /* No declaration, so sniff next token - if it is 'as' then parse
+             * as a routine, otherwise as a grouping.
+             */
+            if (read_token(a->tokeniser) == c_as) {
+                type = t_routine;
+            } else {
+                type = t_grouping;
+            }
+            a->tokeniser->token_held = true;
+        }
+
+        if (type == t_grouping) {
+            read_define_grouping(a, q);
+        } else {
+            read_define_routine(a, q);
+        }
     }
 }
 
@@ -923,6 +1033,7 @@ static void read_program_(struct analyser * a, int terminator) {
             case c_backwardmode:read_backwardmode(a); break;
             case c_ket:
                 if (terminator == c_ket) return;
+                /* fall through */
             default:
                 error(a, e_unexpected_token); break;
             case -1:
@@ -939,9 +1050,11 @@ extern void read_program(struct analyser * a) {
         while (q) {
             switch (q->type) {
                 case t_external: case t_routine:
-                    if (q->used && q->definition == 0) error4(a, q); break;
+                    if (q->used && q->definition == 0) error4(a, q);
+                    break;
                 case t_grouping:
-                    if (q->used && q->grouping == 0) error4(a, q); break;
+                    if (q->used && q->grouping == 0) error4(a, q);
+                    break;
             }
             q = q->next;
         }
@@ -949,33 +1062,56 @@ extern void read_program(struct analyser * a) {
 
     if (a->tokeniser->error_count == 0) {
         struct name * q = a->names;
-        int warned = false;
         while (q) {
             if (!q->referenced) {
-                if (!warned) {
-                    fprintf(stderr, "Declared but not used:");
-                    warned = true;
+                fprintf(stderr, "%s:%d: warning: %s '",
+                        a->tokeniser->file,
+                        q->declaration_line_number,
+                        name_of_name_type(q->type));
+                report_b(stderr, q->b);
+                if (q->type == t_routine ||
+                    q->type == t_external ||
+                    q->type == t_grouping) {
+                    fprintf(stderr, "' declared but not defined\n");
+                } else {
+                    fprintf(stderr, "' defined but not used\n");
                 }
-                fprintf(stderr, " "); report_b(stderr, q->b);
+            } else if (q->type == t_routine || q->type == t_grouping) {
+                if (!q->used) {
+                    int line_num;
+                    if (q->type == t_routine) {
+                        line_num = q->definition->line_number;
+                    } else {
+                        line_num = q->grouping->line_number;
+                    }
+                    fprintf(stderr, "%s:%d: warning: %s '",
+                            a->tokeniser->file,
+                            line_num,
+                            name_of_name_type(q->type));
+                    report_b(stderr, q->b);
+                    fprintf(stderr, "' defined but not used\n");
+                }
+            } else if (q->type == t_external) {
+                /* Unused is OK. */
+            } else if (!q->initialised) {
+                count_error(a);
+                fprintf(stderr, "%s:%d: warning: %s '",
+                        a->tokeniser->file,
+                        q->declaration_line_number,
+                        name_of_name_type(q->type));
+                report_b(stderr, q->b);
+                fprintf(stderr, "' is never initialised\n");
+            } else if (!q->value_used) {
+                count_error(a);
+                fprintf(stderr, "%s:%d: warning: %s '",
+                        a->tokeniser->file,
+                        q->declaration_line_number,
+                        name_of_name_type(q->type));
+                report_b(stderr, q->b);
+                fprintf(stderr, "' is set but never used\n");
             }
             q = q->next;
         }
-        if (warned) fprintf(stderr, "\n");
-
-        q = a->names;
-        warned = false;
-        while (q) {
-            if (! q->used && (q->type == t_routine ||
-                              q->type == t_grouping)) {
-                if (!warned) {
-                    fprintf(stderr, "Declared and defined but not used:");
-                    warned = true;
-                }
-                fprintf(stderr, " "); report_b(stderr, q->b);
-            }
-            q = q->next;
-        }
-        if (warned) fprintf(stderr, "\n");
     }
 }
 
