@@ -83,11 +83,10 @@ static void write_hex(struct generator * g, int i) {
 static void wlitch(struct generator * g, int ch) {
     if (32 <= ch && ch < 127) {
         write_char(g, '\'');
-        switch (ch) {
-            case '\'':
-            case '\\': write_char(g, '\\');
-            default:   write_char(g, ch);
+        if (ch == '\'' || ch == '\\') {
+            write_char(g, '\\');
         }
+        write_char(g, ch);
         write_char(g, '\'');
     }  else {
         write_string(g, "0x"); write_hex(g, ch);
@@ -175,13 +174,29 @@ static void write_block_end(struct generator * g) {    /* block end */
 
 static void w(struct generator * g, const char * s);
 
-static void wk(struct generator * g, struct node * p) {     /* keep c */
+/* keep c */
+static void wk(struct generator * g, struct node * p, int keep_limit) {
     ++g->keep_count;
     if (p->mode == m_forward) {
-        write_string(g, "int c"); write_int(g, g->keep_count); w(g, " = ~zc;");
+        write_string(g, "int c");
+        write_int(g, g->keep_count);
+        w(g, " = ~zc");
+	if (keep_limit) {
+            write_string(g, ", mlimit");
+            write_int(g, g->keep_count);
+        }
+        write_char(g, ';');
     } else {
-        write_string(g, "int m"); write_int(g, g->keep_count); w(g, " = ~zl - ~zc; /*(void)m");
-        write_int(g, g->keep_count); write_string(g, "*/;");
+        write_string(g, "int m");
+        write_int(g, g->keep_count);
+        w(g, " = ~zl - ~zc");
+	if (keep_limit) {
+            write_string(g, ", mlimit");
+            write_int(g, g->keep_count);
+        }
+        write_string(g, "; (void)m");
+        write_int(g, g->keep_count);
+        write_char(g, ';');
     }
 }
 
@@ -267,6 +282,7 @@ static void write_data_address(struct generator * g, struct node * p) {
 static void writef(struct generator * g, const char * input, struct node * p) {
     int i = 0;
     int l = strlen(input);
+
     while (i < l) {
         int ch = input[i++];
         if (ch != '~') {
@@ -276,7 +292,8 @@ static void writef(struct generator * g, const char * input, struct node * p) {
         switch (input[i++]) {
             default: write_char(g, input[i - 1]); continue;
             case 'C': write_comment(g, p); continue;
-            case 'k': wk(g, p); continue;
+            case 'k': wk(g, p, false); continue;
+            case 'K': wk(g, p, true); continue;
             case 'i': winc(g, p); continue;
             case 'l': write_check_limit(g, p); continue;
             case 'f': write_failure(g, p); continue;
@@ -316,7 +333,7 @@ static void w(struct generator * g, const char * s) {
 }
 
 static void generate_AE(struct generator * g, struct node * p) {
-    char * s;
+    const char * s;
     switch (p->type) {
         case c_name:
             write_varref(g, p->name); break;
@@ -344,7 +361,7 @@ static void generate_AE(struct generator * g, struct node * p) {
         case c_limit:
             w(g, p->mode == m_forward ? "~zl" : "~zlb"); break;
         case c_len:
-            if (g->options->utf8) {
+            if (g->options->encoding == ENC_UTF8) {
                 w(g, "len_utf8(~zp)");
                 break;
             }
@@ -353,7 +370,7 @@ static void generate_AE(struct generator * g, struct node * p) {
             w(g, "SIZE(~zp)");
             break;
         case c_lenof:
-            if (g->options->utf8) {
+            if (g->options->encoding == ENC_UTF8) {
                 g->V[0] = p->name;
                 w(g, "len_utf8(~V0)");
                 break;
@@ -367,13 +384,15 @@ static void generate_AE(struct generator * g, struct node * p) {
 }
 
 /* K_needed() tests to see if we really need to keep c. Not true when the
-   the command does not touch the cursor. This and repeat_score() could be
+   command does not touch the cursor. This and repeat_score() could be
    elaborated almost indefinitely.
 */
 
-static int K_needed(struct generator * g, struct node * p) {
+static int K_needed_(struct generator * g, struct node * p, int call_depth) {
     while (p) {
         switch (p->type) {
+            case c_atlimit:
+            case c_do:
             case c_dollar:
             case c_leftslice:
             case c_rightslice:
@@ -390,17 +409,27 @@ static int K_needed(struct generator * g, struct node * p) {
             case c_le:
             case c_sliceto:
             case c_booltest:
+            case c_set:
+            case c_unset:
             case c_true:
             case c_false:
             case c_debug:
                 break;
 
             case c_call:
-                if (K_needed(g, p->name->definition)) return true;
+                /* Recursive functions aren't typical in snowball programs, so
+                 * make the pessimistic assumption that keep is needed if we
+                 * hit a generous limit on recursion.  It's not likely to make
+                 * a difference to any real world program, but means we won't
+                 * recurse until we run out of stack for pathological cases.
+                 */
+                if (call_depth >= 100) return true;
+                if (K_needed_(g, p->name->definition, call_depth + 1))
+                    return true;
                 break;
 
             case c_bra:
-                if (K_needed(g, p->left)) return true;
+                if (K_needed_(g, p->left, call_depth)) return true;
                 break;
 
             default: return true;
@@ -410,7 +439,11 @@ static int K_needed(struct generator * g, struct node * p) {
     return false;
 }
 
-static int repeat_score(struct generator * g, struct node * p) {
+extern int K_needed(struct generator * g, struct node * p) {
+    return K_needed_(g, p, 0);
+}
+
+static int repeat_score(struct generator * g, struct node * p, int call_depth) {
     int score = 0;
     while (p) {
         switch (p->type) {
@@ -433,11 +466,25 @@ static int repeat_score(struct generator * g, struct node * p) {
                 break;
 
             case c_call:
-                score += repeat_score(g, p->name->definition);
+                /* Recursive functions aren't typical in snowball programs, so
+                 * make the pessimistic assumption that repeat requires cursor
+                 * reinstatement if we hit a generous limit on recursion.  It's
+                 * not likely to make a difference to any real world program,
+                 * but means we won't recurse until we run out of stack for
+                 * pathological cases.
+                 */
+                if (call_depth >= 100) {
+                    return 2;
+                }
+                score += repeat_score(g, p->name->definition, call_depth + 1);
+                if (score >= 2)
+                    return score;
                 break;
 
             case c_bra:
-                score += repeat_score(g, p->left);
+                score += repeat_score(g, p->left, call_depth);
+                if (score >= 2)
+                    return score;
                 break;
 
             case c_name:
@@ -446,12 +493,12 @@ static int repeat_score(struct generator * g, struct node * p) {
             case c_grouping:
             case c_non:
             case c_hop:
-                score = score + 1;
+                if (++score >= 2)
+                    return score;
                 break;
 
             default:
-                score = 2;
-                break;
+                return 2;
         }
         p = p->right;
     }
@@ -460,8 +507,8 @@ static int repeat_score(struct generator * g, struct node * p) {
 
 /* tests if an expression requires cursor reinstatement in a repeat */
 
-static int repeat_restore(struct generator * g, struct node * p) {
-    return repeat_score(g, p) >= 2;
+extern int repeat_restore(struct generator * g, struct node * p) {
+    return repeat_score(g, p, 0) >= 2;
 }
 
 static void generate_bra(struct generator * g, struct node * p) {
@@ -531,7 +578,7 @@ static void generate_or(struct generator * g, struct node * p) {
 
 static void generate_backwards(struct generator * g, struct node * p) {
 
-    writef(g,"~M~zlb = ~zc; ~zc = ~zl;~C~N", p);
+    writef(g, "~M~zlb = ~zc; ~zc = ~zl;~C~N", p);
     generate(g, p->left);
     w(g, "~M~zc = ~zlb;~N");
 }
@@ -654,7 +701,7 @@ static void generate_do(struct generator * g, struct node * p) {
 }
 
 static void generate_next(struct generator * g, struct node * p) {
-    if (g->options->utf8) {
+    if (g->options->encoding == ENC_UTF8) {
         if (p->mode == m_forward)
             w(g, "~{int ret = skip_utf8(~zp, ~zc, 0, ~zl, 1");
         else
@@ -673,7 +720,7 @@ static void generate_GO_grouping(struct generator * g, struct node * p, int is_g
     struct grouping * q = p->name->grouping;
     g->S[0] = p->mode == m_forward ? "" : "_b";
     g->S[1] = complement ? "in" : "out";
-    g->S[2] = g->options->utf8 ? "_U" : "";
+    g->S[2] = g->options->encoding == ENC_UTF8 ? "_U" : "";
     g->V[0] = p->name;
     g->I[0] = q->smallest_ch;
     g->I[1] = q->largest_ch;
@@ -817,7 +864,7 @@ static void generate_hop(struct generator * g, struct node * p) {
     g->S[0] = p->mode == m_forward ? "+" : "-";
     g->S[1] = p->mode == m_forward ? "0" :
 	(g->options->make_lang == LANG_C ? "z->lb" : "lb");
-    if (g->options->utf8) {
+    if (g->options->encoding == ENC_UTF8) {
         w(g, "~{int ret = skip_utf8(~zp, ~zc, ~S1, ~zl, ~S0 ");
         generate_AE(g, p->AE); writef(g, ");~C", p);
         writef(g, "~Mif (ret < 0) ~f~N", p);
@@ -919,18 +966,49 @@ static void generate_slicefrom(struct generator * g, struct node * p) {
 
 static void generate_setlimit(struct generator * g, struct node * p) {
     int keep_c;
-    writef(g, "~{~k~C", p);
-    keep_c = g->keep_count;
-    w(g, "~Mint mlimit");
-    write_int(g, keep_c);
-    w(g, ";~N");
-    generate(g, p->left);
+    if (p->left && p->left->type == c_tomark && !p->left->right) {
+        /* Special case for:
+         *
+         *   setlimit tomark AE for C
+         *
+         * All uses of setlimit in the current stemmers we ship follow this
+         * pattern, and by special-casing we can avoid having to save and
+         * restore c.
+         */
+        struct node * q = p->left;
+
+        ++g->keep_count;
+        writef(g, "~N~{int mlimit", p);
+        write_int(g, g->keep_count);
+        writef(g, ";~C", p);
+        keep_c = g->keep_count;
+
+        g->S[0] = q->mode == m_forward ? ">" : "<";
+
+        w(g, "~Mif (~zc ~S0 "); generate_AE(g, q->AE); writef(g, ") ~f~N", q);
+        w(g, "~Mmlimit");
+        write_int(g, keep_c);
+        if (p->mode == m_forward) {
+            w(g, " = ~zl - ~zc; ~zl = ");
+        } else {
+            w(g, " = ~zlb; ~zlb = ");
+        }
+        generate_AE(g, q->AE);
+        w(g, ";~N");
+    } else {
+        writef(g, "~{~K~C", p);
+        keep_c = g->keep_count;
+        generate(g, p->left);
+
+        w(g, "~Mmlimit");
+        write_int(g, keep_c);
+        if (p->mode == m_forward)
+            w(g, " = ~zl - ~zc; ~zl = ~zc;~N");
+        else
+            w(g, " = ~zlb; ~zlb = ~zc;~N");
+        w(g, "~M"); wrestore(g, p, keep_c); w(g, "~N");
+    }
 
-    w(g, "~Mmlimit");
-    write_int(g, keep_c);
-    if (p->mode == m_forward) w(g, " = ~zl - ~zc; ~zl = ~zc;~N");
-                         else w(g, " = ~zlb; ~zlb = ~zc;~N");
-    w(g, "~M"); wrestore(g, p, keep_c); w(g, "~N");
     g->failure_keep_count = -keep_c;
     generate(g, p->aux);
     w(g, "~M");
@@ -941,6 +1019,8 @@ static void generate_setlimit(struct generator * g, struct node * p) {
 
 static const char * vars[] = { "p", "c", "l", "lb", "bra", "ket", NULL };
 
+/* dollar sets snowball up to operate on a string variable as if it were the
+ * current string */
 static void generate_dollar(struct generator * g, struct node * p) {
 
     int used = g->label_used;
@@ -1027,7 +1107,7 @@ static void generate_grouping(struct generator * g, struct node * p, int complem
     struct grouping * q = p->name->grouping;
     g->S[0] = p->mode == m_forward ? "" : "_b";
     g->S[1] = complement ? "out" : "in";
-    g->S[2] = g->options->utf8 ? "_U" : "";
+    g->S[2] = g->options->encoding == ENC_UTF8 ? "_U" : "";
     g->V[0] = p->name;
     g->I[0] = q->smallest_ch;
     g->I[1] = q->largest_ch;
@@ -1047,9 +1127,9 @@ static void generate_literalstring(struct generator * g, struct node * p) {
         /* It's quite common to compare with a single character literal string,
          * so just inline the simpler code for this case rather than making a
          * function call.  In UTF-8 mode, only do this for the ASCII subset,
-         * since multi-byte characters are more complex to text against.
+         * since multi-byte characters are more complex to test against.
          */
-        if (g->options->utf8 && *b >= 128) {
+        if (g->options->encoding == ENC_UTF8 && *b >= 128) {
             printf("single byte %d\n", *b);
             exit(1);
         }
@@ -1144,7 +1224,7 @@ static void generate_substring(struct generator * g, struct node * p) {
             if (n_cases > 2) break;
         }
         if (block == -1) {
-            if (ch == cases[0]) continue;
+            if (n_cases > 0 && ch == cases[0]) continue;
             if (n_cases < 2) {
                 cases[n_cases++] = ch;
             } else if (ch != cases[1]) {
@@ -1248,7 +1328,7 @@ static void generate_among(struct generator * g, struct node * p) {
     if (x->substring == 0) generate_substring(g, p);
     if (x->command_count == 0 && x->starter == 0) return;
 
-    if (x->starter) generate(g, x->starter);
+    if (x->starter != 0) generate(g, x->starter);
 
     writef(g, "~Mswitch (among_var) {~C~+"
               "~Mcase 0: ~f~N", p);
@@ -1586,16 +1666,14 @@ static void generate_grouping_table(struct generator * g, struct grouping * q) {
 
     for (i = 0; i < SIZE(b); i++) set_bit(map, b[i] - q->smallest_ch);
 
-    {
-        g->V[0] = q->name;
+    g->V[0] = q->name;
 
-        w(g, "static const unsigned char ~V0[] = { ");
-        for (i = 0; i < size; i++) {
-             write_int(g, map[i]);
-             if (i < size - 1) w(g, ", ");
-        }
-        w(g, " };~N~N");
+    w(g, "static const unsigned char ~V0[] = { ");
+    for (i = 0; i < size; i++) {
+        write_int(g, map[i]);
+        if (i < size - 1) w(g, ", ");
     }
+    w(g, " };~N~N");
     lose_b(map);
 }
 
