@@ -1,3 +1,4 @@
+#include <stdio.h>
 
 typedef unsigned char byte;
 typedef unsigned short symbol;
@@ -47,6 +48,8 @@ extern int get_utf8(const symbol * p, int * slot);
 extern int put_utf8(int ch, symbol * p);
 extern void output_str(FILE * outfile, struct str * str);
 
+typedef enum { ENC_SINGLEBYTE, ENC_UTF8, ENC_WIDECHARS } enc;
+
 struct m_pair {
 
     struct m_pair * next;
@@ -89,6 +92,12 @@ enum token_codes {
     NUM_TOKEN_CODES
 };
 
+enum uplus_modes {
+    UPLUS_NONE,
+    UPLUS_DEFINED,
+    UPLUS_UNICODE
+};
+
 /* struct input must be a prefix of struct tokeniser. */
 struct tokeniser {
 
@@ -108,12 +117,18 @@ struct tokeniser {
     int token;
     int previous_token;
     byte token_held;
-    byte widechars;
-    byte utf8;
+    enc encoding;
 
     int omission;
     struct include * includes;
 
+    /* Mode in which U+ has been used:
+     * UPLUS_NONE - not used yet
+     * UPLUS_DEFINED - stringdef U+xxxx ....
+     * UPLUS_UNICODE - {U+xxxx} used with implicit meaning
+     */
+    int uplusmode;
+
     char token_disabled[NUM_TOKEN_CODES];
 };
 
@@ -142,8 +157,11 @@ struct name {
     struct grouping * grouping; /* for grouping names */
     byte referenced;
     byte used_in_among;         /* Function used in among? */
+    byte value_used;            /* (For variables) is its value ever used? */
+    byte initialised;           /* (For variables) is it ever initialised? */
     struct node * used;         /* First use, or NULL if not used */
     struct name * local_to;     /* Local to one routine/external */
+    int declaration_line_number;/* Line number of declaration */
 
 };
 
@@ -185,6 +203,7 @@ struct grouping {
     int largest_ch;           /* character with max code */
     int smallest_ch;          /* character with min code */
     struct name * name;       /* so g->name->grouping == g */
+    int line_number;
 };
 
 struct node {
@@ -243,7 +262,7 @@ struct analyser {
     struct grouping * groupings;
     struct grouping * groupings_end;
     struct node * substring;  /* pending 'substring' in current routine definition */
-    byte utf8;
+    enc encoding;
     byte int_limits_used;     /* are maxint or minint used? */
 };
 
@@ -278,7 +297,7 @@ struct generator {
      * if < 0, the negated keep_count for the limit to restore in case of
      * failure. */
     int failure_keep_count;
-#if !defined(DISABLE_JAVA) && !defined(DISABLE_JSX) && !defined(DISABLE_PYTHON)
+#if !defined(DISABLE_JAVA) && !defined(DISABLE_JS) && !defined(DISABLE_PYTHON) && !defined(DISABLE_CSHARP)
     struct str * failure_str;  /* This is used by some generators instead of failure_keep_count */
 #endif
 
@@ -309,18 +328,19 @@ struct options {
     FILE * output_src;
     FILE * output_h;
     byte syntax_tree;
-    byte widechars;
-    enum { LANG_JAVA, LANG_C, LANG_CPLUSPLUS, LANG_PYTHON, LANG_JSX } make_lang;
+    enc encoding;
+    enum { LANG_JAVA, LANG_C, LANG_CPLUSPLUS, LANG_CSHARP, LANG_PYTHON, LANG_JAVASCRIPT, LANG_RUST, LANG_GO } make_lang;
     const char * externals_prefix;
     const char * variables_prefix;
     const char * runtime_path;
     const char * parent_class_name;
     const char * package;
+    const char * go_package;
+    const char * go_snowball_runtime;
     const char * string_class;
     const char * among_class;
     struct include * includes;
     struct include * includes_end;
-    byte utf8;
 };
 
 /* Generator functions common to several backends. */
@@ -335,6 +355,9 @@ extern void write_int(struct generator * g, int i);
 extern void write_b(struct generator * g, symbol * b);
 extern void write_str(struct generator * g, struct str * str);
 
+extern int K_needed(struct generator * g, struct node * p);
+extern int repeat_restore(struct generator * g, struct node * p);
+
 /* Generator for C code. */
 extern void generate_program_c(struct generator * g);
 
@@ -343,11 +366,24 @@ extern void generate_program_c(struct generator * g);
 extern void generate_program_java(struct generator * g);
 #endif
 
+#ifndef DISABLE_CSHARP
+/* Generator for C# code. */
+extern void generate_program_csharp(struct generator * g);
+#endif
+
 #ifndef DISABLE_PYTHON
 /* Generator for Python code. */
 extern void generate_program_python(struct generator * g);
 #endif
 
-#ifndef DISABLE_JSX
-extern void generate_program_jsx(struct generator * g);
+#ifndef DISABLE_JS
+extern void generate_program_js(struct generator * g);
+#endif
+
+#ifndef DISABLE_RUST
+extern void generate_program_rust(struct generator * g);
+#endif
+
+#ifndef DISABLE_GO
+extern void generate_program_go(struct generator * g);
 #endif
