@@ -28,9 +28,6 @@
 
 class LeafPostList;
 class MultiMatch;
-namespace Xapian {
-class Weight;
-}
 
 class QueryOptimiser {
     /// Prevent assignment.
@@ -41,7 +38,7 @@ class QueryOptimiser {
 
     LocalSubMatch & localsubmatch;
 
-    /** How many leaf subqueries there are.
+    /** How many weighted leaf subqueries there are.
      *
      *  Used for scaling percentages when the highest weighted document doesn't
      *  "match all terms".
