@@ -1,7 +1,7 @@
 /** @file  remoteconnection.cc
  *  @brief RemoteConnection class used by the remote backend.
  */
-/* Copyright (C) 2006,2007,2008,2009,2010,2011,2014,2015 Olly Betts
+/* Copyright (C) 2006,2007,2008,2009,2010,2011,2012,2013,2014,2015,2017 Olly Betts
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -27,30 +27,37 @@
 #include "safeerrno.h"
 #include "safefcntl.h"
 #include "safesysselect.h"
-#include "safesysstat.h"
 #include "safeunistd.h"
 
 #include <algorithm>
 #include <climits>
+#include <cstdint>
 #include <string>
+#ifdef __WIN32__
+# include <type_traits>
+#endif
 
 #include "debuglog.h"
-#include "net/length.h"
+#include "fd.h"
+#include "filetests.h"
 #include "noreturn.h"
 #include "omassert.h"
+#include "posixy_wrapper.h"
 #include "realtime.h"
-#include "serialise.h"
+#include "length.h"
 #include "socket_utils.h"
-#include "utils.h"
-
-#ifdef __WIN32__
-# include "msvc_posix_wrapper.h"
-#endif
 
 using namespace std;
 
 #define CHUNKSIZE 4096
 
+XAPIAN_NORETURN(static void throw_database_closed());
+static void
+throw_database_closed()
+{
+    throw Xapian::DatabaseError("Database has been closed");
+}
+
 XAPIAN_NORETURN(static void throw_network_error_insane_message_length());
 static void
 throw_network_error_insane_message_length()
@@ -62,7 +69,8 @@ throw_network_error_insane_message_length()
 inline void
 update_overlapped_offset(WSAOVERLAPPED & overlapped, DWORD n)
 {
-    STATIC_ASSERT_UNSIGNED_TYPE(DWORD); // signed overflow is undefined.
+    // Signed overflow is undefined so check DWORD is unsigned.
+    static_assert(std::is_unsigned<DWORD>::value, "Type DWORD should be unsigned");
     overlapped.Offset += n;
     if (overlapped.Offset < n) ++overlapped.OffsetHigh;
 }
@@ -77,18 +85,18 @@ RemoteConnection::RemoteConnection(int fdin_, int fdout_,
     overlapped.hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
     if (!overlapped.hEvent)
 	throw Xapian::NetworkError("Failed to setup OVERLAPPED",
-				   context, -(int)GetLastError());
+				   context, -int(GetLastError()));
 
 #endif
 }
 
+#ifdef __WIN32__
 RemoteConnection::~RemoteConnection()
 {
-#ifdef __WIN32__
     if (overlapped.hEvent)
 	CloseHandle(overlapped.hEvent);
-#endif
 }
+#endif
 
 bool
 RemoteConnection::read_at_least(size_t min_len, double end_time)
@@ -117,7 +125,7 @@ RemoteConnection::read_at_least(size_t min_len, double end_time)
 	    // Get the final result of the read.
 	    if (!GetOverlappedResult(hin, &overlapped, &received, FALSE))
 		throw Xapian::NetworkError("Failed to get overlapped result",
-					   context, -(int)GetLastError());
+					   context, -int(GetLastError()));
 	}
 
 	if (received == 0) {
@@ -169,22 +177,25 @@ RemoteConnection::read_at_least(size_t min_len, double end_time)
 	    }
 
 	    // Use select to wait until there is data or the timeout is reached.
-	    fd_set fdset;
-	    FD_ZERO(&fdset);
-	    FD_SET(fdin, &fdset);
+	    fd_set rfdset, efdset;
+	    FD_ZERO(&rfdset);
+	    FD_SET(fdin, &rfdset);
+	    FD_ZERO(&efdset);
+	    FD_SET(fdin, &efdset);
 
 	    struct timeval tv;
-	    tv.tv_sec = long(time_diff);
-	    tv.tv_usec = long(fmod(time_diff, 1.0) * 1000000);
-
-	    int select_result = select(fdin + 1, &fdset, 0, &fdset, &tv);
+	    RealTime::to_timeval(time_diff, &tv);
+	    int select_result = select(fdin + 1, &rfdset, 0, &efdset, &tv);
 	    if (select_result > 0) break;
 
 	    if (select_result == 0)
 		throw Xapian::NetworkTimeoutError("Timeout expired while trying to read", context);
 
-	    // EINTR means select was interrupted by a signal.
-	    if (errno != EINTR)
+	    // EINTR means select was interrupted by a signal.  The Linux
+	    // select(2) man page says: "Portable programs may wish to check
+	    // for EAGAIN and loop, just as with EINTR" and that seems to be
+	    // necessary for cygwin at least.
+	    if (errno != EINTR && errno != EAGAIN)
 		throw Xapian::NetworkError("select failed during read", context, errno);
 	}
     }
@@ -196,16 +207,17 @@ bool
 RemoteConnection::ready_to_read() const
 {
     LOGCALL(REMOTE, bool, "RemoteConnection::ready_to_read", NO_ARGS);
-    if (fdin == -1) {
-	throw Xapian::DatabaseError("Database has been closed");
-    }
+    if (fdin == -1)
+	throw_database_closed();
 
     if (!buffer.empty()) RETURN(true);
 
     // Use select to see if there's data available to be read.
-    fd_set fdset;
-    FD_ZERO(&fdset);
-    FD_SET(fdin, &fdset);
+    fd_set rfdset, efdset;
+    FD_ZERO(&rfdset);
+    FD_SET(fdin, &rfdset);
+    FD_ZERO(&efdset);
+    FD_SET(fdin, &efdset);
 
     // Set a 0.1 second timeout to avoid a busy loop.
     // FIXME: this would be much better done by exposing the fd so that the
@@ -213,7 +225,7 @@ RemoteConnection::ready_to_read() const
     struct timeval tv;
     tv.tv_sec = 0;
     tv.tv_usec = 100000;
-    RETURN(select(fdin + 1, &fdset, 0, &fdset, &tv) > 0);
+    RETURN(select(fdin + 1, &rfdset, 0, &efdset, &tv) > 0);
 }
 
 void
@@ -221,9 +233,8 @@ RemoteConnection::send_message(char type, const string &message,
 			       double end_time)
 {
     LOGCALL_VOID(REMOTE, "RemoteConnection::send_message", type | message | end_time);
-    if (fdout == -1) {
-	throw Xapian::DatabaseError("Database has been closed");
-    }
+    if (fdout == -1)
+	throw_database_closed();
 
     string header;
     header += type;
@@ -251,7 +262,7 @@ RemoteConnection::send_message(char type, const string &message,
 	    // Get the final result.
 	    if (!GetOverlappedResult(hout, &overlapped, &n, FALSE))
 		throw Xapian::NetworkError("Failed to get overlapped result",
-					   context, -(int)GetLastError());
+					   context, -int(GetLastError()));
 	}
 
 	count += n;
@@ -274,7 +285,7 @@ RemoteConnection::send_message(char type, const string &message,
 
     const string * str = &header;
 
-    fd_set fdset;
+    fd_set wfdset, efdset;
     size_t count = 0;
     while (true) {
 	// We've set write to non-blocking, so just try writing as there
@@ -298,8 +309,10 @@ RemoteConnection::send_message(char type, const string &message,
 	    throw Xapian::NetworkError("write failed", context, errno);
 
 	// Use select to wait until there is space or the timeout is reached.
-	FD_ZERO(&fdset);
-	FD_SET(fdout, &fdset);
+	FD_ZERO(&wfdset);
+	FD_SET(fdout, &wfdset);
+	FD_ZERO(&efdset);
+	FD_SET(fdout, &efdset);
 
 	double time_diff = end_time - RealTime::now();
 	if (time_diff < 0) {
@@ -308,10 +321,8 @@ RemoteConnection::send_message(char type, const string &message,
 	}
 
 	struct timeval tv;
-	tv.tv_sec = long(time_diff);
-	tv.tv_usec = long(fmod(time_diff, 1.0) * 1000000);
-
-	int select_result = select(fdout + 1, 0, &fdset, &fdset, &tv);
+	RealTime::to_timeval(time_diff, &tv);
+	int select_result = select(fdout + 1, 0, &wfdset, &efdset, &tv);
 
 	if (select_result < 0) {
 	    if (errno == EINTR) {
@@ -333,17 +344,12 @@ void
 RemoteConnection::send_file(char type, int fd, double end_time)
 {
     LOGCALL_VOID(REMOTE, "RemoteConnection::send_file", type | fd | end_time);
-    if (fdout == -1) {
-	throw Xapian::DatabaseError("Database has been closed");
-    }
+    if (fdout == -1)
+	throw_database_closed();
 
-    off_t size;
-    {
-	struct stat sb;
-	if (fstat(fd, &sb) == -1)
-	    throw Xapian::NetworkError("Couldn't stat file to send", errno);
-	size = sb.st_size;
-    }
+    off_t size = file_size(fd);
+    if (errno)
+	throw Xapian::NetworkError("Couldn't stat file to send", errno);
     // FIXME: Use sendfile() or similar if available?
 
     char buf[CHUNKSIZE];
@@ -377,7 +383,7 @@ RemoteConnection::send_file(char type, int fd, double end_time)
 	    // Get the final result.
 	    if (!GetOverlappedResult(hout, &overlapped, &n, FALSE))
 		throw Xapian::NetworkError("Failed to get overlapped result",
-					   context, -(int)GetLastError());
+					   context, -int(GetLastError()));
 	}
 
 	count += n;
@@ -406,7 +412,7 @@ RemoteConnection::send_file(char type, int fd, double end_time)
 				   context, errno);
     }
 
-    fd_set fdset;
+    fd_set wfdset, efdset;
     size_t count = 0;
     while (true) {
 	// We've set write to non-blocking, so just try writing as there
@@ -438,8 +444,10 @@ RemoteConnection::send_file(char type, int fd, double end_time)
 	    throw Xapian::NetworkError("write failed", context, errno);
 
 	// Use select to wait until there is space or the timeout is reached.
-	FD_ZERO(&fdset);
-	FD_SET(fdout, &fdset);
+	FD_ZERO(&wfdset);
+	FD_SET(fdout, &wfdset);
+	FD_ZERO(&efdset);
+	FD_SET(fdout, &efdset);
 
 	double time_diff = end_time - RealTime::now();
 	if (time_diff < 0) {
@@ -448,10 +456,8 @@ RemoteConnection::send_file(char type, int fd, double end_time)
 	}
 
 	struct timeval tv;
-	tv.tv_sec = long(time_diff);
-	tv.tv_usec = long(fmod(time_diff, 1.0) * 1000000);
-
-	int select_result = select(fdout + 1, 0, &fdset, &fdset, &tv);
+	RealTime::to_timeval(time_diff, &tv);
+	int select_result = select(fdout + 1, 0, &wfdset, &efdset, &tv);
 
 	if (select_result < 0) {
 	    if (errno == EINTR) {
@@ -473,9 +479,8 @@ int
 RemoteConnection::sniff_next_message_type(double end_time)
 {
     LOGCALL(REMOTE, int, "RemoteConnection::sniff_next_message_type", end_time);
-    if (fdin == -1) {
-	throw Xapian::DatabaseError("Database has been closed");
-    }
+    if (fdin == -1)
+	throw_database_closed();
 
     if (!read_at_least(1, end_time))
 	RETURN(-1);
@@ -487,9 +492,8 @@ int
 RemoteConnection::get_message(string &result, double end_time)
 {
     LOGCALL(REMOTE, int, "RemoteConnection::get_message", result | end_time);
-    if (fdin == -1) {
-	throw Xapian::DatabaseError("Database has been closed");
-    }
+    if (fdin == -1)
+	throw_database_closed();
 
     if (!read_at_least(2, end_time))
 	RETURN(-1);
@@ -529,17 +533,15 @@ int
 RemoteConnection::get_message_chunked(double end_time)
 {
     LOGCALL(REMOTE, int, "RemoteConnection::get_message_chunked", end_time);
-    typedef UNSIGNED_OFF_T uoff_t;
 
-    if (fdin == -1) {
-	throw Xapian::DatabaseError("Database has been closed");
-    }
+    if (fdin == -1)
+	throw_database_closed();
 
     if (!read_at_least(2, end_time))
 	RETURN(-1);
-    uoff_t len = static_cast<unsigned char>(buffer[1]);
+    uint_least64_t len = static_cast<unsigned char>(buffer[1]);
     if (len != 0xff) {
-	chunked_data_left = len;
+	chunked_data_left = off_t(len);
 	char type = buffer[0];
 	buffer.erase(0, 2);
 	RETURN(type);
@@ -554,37 +556,28 @@ RemoteConnection::get_message_chunked(double end_time)
 	// Allow at most 63 bits for message lengths - it's neatly a multiple
 	// of 7 bits and anything longer than this is almost certainly a
 	// corrupt value.
-#if SIZEOF_OFF_T < 8
-	// The value also needs to be representable as an off_t, which is a
-	// signed type.
-	if (rare(i == buffer.end() || shift >= SIZEOF_OFF_T * CHAR_BIT - 1)) {
+	// The value also needs to be representable as an
+	// off_t (which is a signed type), so use that size if it is smaller.
+	const int SHIFT_LIMIT = 63;
+	if (rare(i == buffer.end() || shift >= SHIFT_LIMIT)) {
 	    // Something is very wrong...
 	    throw_network_error_insane_message_length();
 	}
-#else
-	if (rare(i == buffer.end() || shift >= 63)) {
-	    // Something is very wrong...
-	    throw_network_error_insane_message_length();
-	}
-#endif
 	ch = *i++;
-	uoff_t bits = ch & 0x7f;
-#if SIZEOF_OFF_T < 8
-	if (shift > (SIZEOF_OFF_T * CHAR_BIT - 7)) {
-	    if (bits >> (shift - (SIZEOF_OFF_T * CHAR_BIT - 7))) {
-		// Too large for off_t.
-		throw_network_error_insane_message_length();
-	    }
-	}
-#endif
+	uint_least64_t bits = ch & 0x7f;
 	len |= bits << shift;
 	shift += 7;
     } while ((ch & 0x80) == 0);
     len += 255;
-    if (rare(off_t(len) < 0))
-	throw_network_error_insane_message_length();
 
-    chunked_data_left = len;
+    chunked_data_left = off_t(len);
+    if (sizeof(off_t) * CHAR_BIT < 63) {
+	// Check that the value of len fits in an off_t without loss.
+	if (rare(uint_least64_t(chunked_data_left) != len)) {
+	    throw_network_error_insane_message_length();
+	}
+    }
+
     unsigned char type = buffer[0];
     size_t header_len = (i - buffer.begin());
     buffer.erase(0, header_len);
@@ -596,9 +589,8 @@ RemoteConnection::get_message_chunk(string &result, size_t at_least,
 				    double end_time)
 {
     LOGCALL(REMOTE, int, "RemoteConnection::get_message_chunk", result | at_least | end_time);
-    if (fdin == -1) {
-	throw Xapian::DatabaseError("Database has been closed");
-    }
+    if (fdin == -1)
+	throw_database_closed();
 
     if (at_least <= result.size()) RETURN(true);
     at_least -= result.size();
@@ -636,18 +628,13 @@ int
 RemoteConnection::receive_file(const string &file, double end_time)
 {
     LOGCALL(REMOTE, int, "RemoteConnection::receive_file", file | end_time);
-    if (fdin == -1) {
-	throw Xapian::DatabaseError("Database has been closed");
-    }
+    if (fdin == -1)
+	throw_database_closed();
 
-#ifdef __WIN32__
-    // Do we want to be able to delete the file during writing?
-    int fd = msvc_posix_open(file.c_str(), O_WRONLY|O_CREAT|O_TRUNC|O_BINARY);
-#else
-    int fd = open(file.c_str(), O_WRONLY|O_CREAT|O_TRUNC|O_BINARY, 0666);
-#endif
-    if (fd == -1) throw Xapian::NetworkError("Couldn't open file for writing: " + file, errno);
-    fdcloser closefd(fd);
+    // FIXME: Do we want to be able to delete the file during writing?
+    FD fd(posixy_open(file.c_str(), O_WRONLY|O_CREAT|O_TRUNC|O_CLOEXEC, 0666));
+    if (fd == -1)
+	throw Xapian::NetworkError("Couldn't open file for writing: " + file, errno);
 
     int type = get_message_chunked(end_time);
     do {
@@ -686,12 +673,14 @@ RemoteConnection::do_close(bool wait)
 #else
 	    // Wait for the connection to be closed - when this happens
 	    // select() will report that a read won't block.
-	    fd_set fdset;
-	    FD_ZERO(&fdset);
-	    FD_SET(fdin, &fdset);
+	    fd_set rfdset, efdset;
+	    FD_ZERO(&rfdset);
+	    FD_SET(fdin, &rfdset);
+	    FD_ZERO(&efdset);
+	    FD_SET(fdin, &efdset);
 	    int res;
 	    do {
-		res = select(fdin + 1, &fdset, 0, &fdset, NULL);
+		res = select(fdin + 1, &rfdset, 0, &efdset, NULL);
 	    } while (res < 0 && errno == EINTR);
 #endif
 	}
