@@ -1,7 +1,7 @@
 /** @file api_compact.cc
  * @brief Tests of Database::compact()
  */
-/* Copyright (C) 2009,2010,2011,2012,2013,2015,2016,2017 Olly Betts
+/* Copyright (C) 2009,2010,2011,2012,2013,2015,2016,2017,2018 Olly Betts
  * Copyright (C) 2010 Richard Boulton
  *
  * This program is free software; you can redistribute it and/or
@@ -29,6 +29,7 @@
 #include "apitest.h"
 #include "dbcheck.h"
 #include "filetests.h"
+#include "msvcignoreinvalidparam.h"
 #include "str.h"
 #include "testsuite.h"
 #include "testutils.h"
@@ -99,7 +100,7 @@ check_sparse_uid_terms(const string & path)
     }
 }
 
-DEFINE_TESTCASE(compactnorenumber1, generated) {
+DEFINE_TESTCASE(compactnorenumber1, compact && generated) {
     string a = get_database_path("compactnorenumber1a", make_sparse_db,
 				 "5-7 24 76 987 1023-1027 9999 !9999");
     string a_uuid;
@@ -236,7 +237,7 @@ DEFINE_TESTCASE(compactnorenumber1, generated) {
 }
 
 // Test use of compact to merge two databases.
-DEFINE_TESTCASE(compactmerge1, chert || glass) {
+DEFINE_TESTCASE(compactmerge1, compact) {
     string indbpath = get_database_path("apitest_simpledata");
     string outdbpath = get_named_writable_database_path("compactmerge1out");
     rm_rf(outdbpath);
@@ -254,15 +255,30 @@ DEFINE_TESTCASE(compactmerge1, chert || glass) {
     TEST_EQUAL(indb.get_doccount() * 2, outdb.get_doccount());
     dbcheck(outdb, outdb.get_doccount(), outdb.get_doccount());
 
-    for (const char * suffix :
-	    { "", "/postlist", "/termlist.", "/docdata.glass" }) {
-	tout.str(string());
-	string arg = outdbpath;
-	if (endswith(suffix, ".glass") && get_dbtype() != "glass") {
-	    suffix = "/record.DB";
+    if (file_exists(outdbpath)) {
+	// Single file case.
+	TEST_EQUAL(Xapian::Database::check(outdbpath, 0, &tout), 0);
+    } else {
+	static const char* const suffixes[] = {
+	    "", "/postlist", "/termlist.", nullptr
+	};
+	for (auto s : suffixes) {
+	    string suffix;
+	    if (s) {
+		suffix = s;
+	    } else {
+		if (get_dbtype() == "chert") {
+		    suffix = "/record.DB";
+		} else {
+		    suffix = "/docdata." + get_dbtype();
+		}
+	    }
+	    tout.str(string());
+	    tout << "Trying suffix '" << suffix << "'" << endl;
+	    string arg = outdbpath;
+	    arg += suffix;
+	    TEST_EQUAL(Xapian::Database::check(arg, 0, &tout), 0);
 	}
-	arg += suffix;
-	TEST_EQUAL(Xapian::Database::check(arg, 0, &tout), 0);
     }
 
     return true;
@@ -285,7 +301,7 @@ make_multichunk_db(Xapian::WritableDatabase &db, const string &)
 
 // Test use of compact on a database which has multiple chunks for a term.
 // This is a regression test for ticket #427
-DEFINE_TESTCASE(compactmultichunks1, generated) {
+DEFINE_TESTCASE(compactmultichunks1, compact && generated) {
     string indbpath = get_database_path("compactmultichunks1in",
 					make_multichunk_db, "");
     string outdbpath = get_named_writable_database_path("compactmultichunks1out");
@@ -306,7 +322,7 @@ DEFINE_TESTCASE(compactmultichunks1, generated) {
 }
 
 // Test compacting from a stub database directory.
-DEFINE_TESTCASE(compactstub1, chert || glass) {
+DEFINE_TESTCASE(compactstub1, compact) {
     const char * stubpath = ".stub/compactstub1";
     const char * stubpathfile = ".stub/compactstub1/XAPIANDB";
     mkdir(".stub", 0755);
@@ -335,7 +351,7 @@ DEFINE_TESTCASE(compactstub1, chert || glass) {
 }
 
 // Test compacting from a stub database file.
-DEFINE_TESTCASE(compactstub2, chert || glass) {
+DEFINE_TESTCASE(compactstub2, compact) {
     const char * stubpath = ".stub/compactstub2";
     mkdir(".stub", 0755);
     ofstream stub(stubpath);
@@ -362,7 +378,7 @@ DEFINE_TESTCASE(compactstub2, chert || glass) {
 }
 
 // Test compacting a stub database file to itself.
-DEFINE_TESTCASE(compactstub3, chert || glass) {
+DEFINE_TESTCASE(compactstub3, compact) {
     const char * stubpath = ".stub/compactstub3";
     mkdir(".stub", 0755);
     ofstream stub(stubpath);
@@ -387,7 +403,7 @@ DEFINE_TESTCASE(compactstub3, chert || glass) {
 }
 
 // Test compacting a stub database directory to itself.
-DEFINE_TESTCASE(compactstub4, chert || glass) {
+DEFINE_TESTCASE(compactstub4, compact) {
     const char * stubpath = ".stub/compactstub4";
     const char * stubpathfile = ".stub/compactstub4/XAPIANDB";
     mkdir(".stub", 0755);
@@ -436,7 +452,7 @@ make_missing_tables(Xapian::WritableDatabase &db, const string &)
     db.commit();
 }
 
-DEFINE_TESTCASE(compactmissingtables1, generated) {
+DEFINE_TESTCASE(compactmissingtables1, compact && generated) {
     string a = get_database_path("compactmissingtables1a",
 				 make_all_tables);
     string b = get_database_path("compactmissingtables1b",
@@ -478,7 +494,7 @@ make_all_tables2(Xapian::WritableDatabase &db, const string &)
 }
 
 /// Adds coverage for merging synonym table.
-DEFINE_TESTCASE(compactmergesynonym1, generated) {
+DEFINE_TESTCASE(compactmergesynonym1, compact && generated) {
     string a = get_database_path("compactmergesynonym1a",
 				 make_all_tables);
     string b = get_database_path("compactmergesynonym1b",
@@ -525,7 +541,7 @@ DEFINE_TESTCASE(compactmergesynonym1, generated) {
     return true;
 }
 
-DEFINE_TESTCASE(compactempty1, chert || glass) {
+DEFINE_TESTCASE(compactempty1, compact) {
     string empty_dbpath = get_database_path(string());
     string outdbpath = get_named_writable_database_path("compactempty1out");
     rm_rf(outdbpath);
@@ -556,7 +572,7 @@ DEFINE_TESTCASE(compactempty1, chert || glass) {
     return true;
 }
 
-DEFINE_TESTCASE(compactmultipass1, chert || glass) {
+DEFINE_TESTCASE(compactmultipass1, compact && generated) {
     string outdbpath = get_named_writable_database_path("compactmultipass1");
     rm_rf(outdbpath);
 
@@ -585,18 +601,25 @@ DEFINE_TESTCASE(compactmultipass1, chert || glass) {
 }
 
 // Test compacting to an fd.
-DEFINE_TESTCASE(compacttofd1, glass) {
+// Chert doesn't support single file databases.
+DEFINE_TESTCASE(compacttofd1, compact && !chert) {
     Xapian::Database indb(get_database("apitest_simpledata"));
     string outdbpath = get_named_writable_database_path("compacttofd1out");
     rm_rf(outdbpath);
 
-    int fd = open(outdbpath.c_str(), O_CREAT|O_RDWR, 0666);
+    int fd = open(outdbpath.c_str(), O_CREAT|O_RDWR|O_BINARY, 0666);
     TEST(fd != -1);
     indb.compact(fd);
 
-    // Confirm that the fd was closed by Xapian.
-    TEST(close(fd) == -1);
-    TEST(errno == EBADF);
+    // Confirm that the fd was closed by Xapian.  Set errno first to workaround
+    // a bug in Wine's msvcrt.dll which fails to set errno in this case:
+    // https://bugs.winehq.org/show_bug.cgi?id=43902
+    errno = EBADF;
+    {
+	MSVCIgnoreInvalidParameter invalid_fd_in_close_is_expected;
+	TEST(close(fd) == -1);
+	TEST_EQUAL(errno, EBADF);
+    }
 
     Xapian::Database outdb(outdbpath);
 
@@ -607,21 +630,28 @@ DEFINE_TESTCASE(compacttofd1, glass) {
 }
 
 // Test compacting to an fd at at offset.
-DEFINE_TESTCASE(compacttofd2, glass) {
+// Chert doesn't support single file databases.
+DEFINE_TESTCASE(compacttofd2, compact && !chert) {
     Xapian::Database indb(get_database("apitest_simpledata"));
     string outdbpath = get_named_writable_database_path("compacttofd2out");
     rm_rf(outdbpath);
 
-    int fd = open(outdbpath.c_str(), O_CREAT|O_RDWR, 0666);
+    int fd = open(outdbpath.c_str(), O_CREAT|O_RDWR|O_BINARY, 0666);
     TEST(fd != -1);
     TEST(lseek(fd, 8192, SEEK_SET) == 8192);
     indb.compact(fd);
 
-    // Confirm that the fd was closed by Xapian.
-    TEST(close(fd) == -1);
-    TEST(errno == EBADF);
+    // Confirm that the fd was closed by Xapian.  Set errno first to workaround
+    // a bug in Wine's msvcrt.dll which fails to set errno in this case:
+    // https://bugs.winehq.org/show_bug.cgi?id=43902
+    errno = EBADF;
+    {
+	MSVCIgnoreInvalidParameter invalid_fd_in_close_is_expected;
+	TEST(close(fd) == -1);
+	TEST_EQUAL(errno, EBADF);
+    }
 
-    fd = open(outdbpath.c_str(), O_RDONLY, 0666);
+    fd = open(outdbpath.c_str(), O_RDONLY|O_BINARY, 0666);
     TEST(fd != -1);
 
     // Test that the database wasn't just written to the start of the file.
@@ -645,9 +675,11 @@ DEFINE_TESTCASE(compacttofd2, glass) {
     return true;
 }
 
-// Regression test for bug fixed in 1.3.5.  If you commit a WritableDatabase
+// Regression test for bug fixed in 1.3.5.  If you compact a WritableDatabase
 // with uncommitted changes, you get an inconsistent output.
-DEFINE_TESTCASE(compactsingle1, glass) {
+//
+// Chert doesn't support single file databases.
+DEFINE_TESTCASE(compactsingle1, compact && !chert) {
     Xapian::WritableDatabase db = get_writable_database();
     Xapian::Document doc;
     doc.add_term("foo");
@@ -655,7 +687,7 @@ DEFINE_TESTCASE(compactsingle1, glass) {
     doc.add_term("baz");
     db.add_document(doc);
 
-    string output = ".glass/db__compactsingle1-out";
+    string output = get_named_writable_database_path("compactsingle1-out");
     // In 1.3.4, we would hang if the output file already existed, so check
     // that works.
     touch(output);
@@ -674,3 +706,28 @@ DEFINE_TESTCASE(compactsingle1, glass) {
 
     return true;
 }
+
+// Regression test for bug fixed in 1.4.6.  Same as above, except not with
+// a single file database!
+DEFINE_TESTCASE(compact1, compact) {
+    Xapian::WritableDatabase db = get_writable_database();
+    Xapian::Document doc;
+    doc.add_term("foo");
+    doc.add_term("bar");
+    doc.add_term("baz");
+    db.add_document(doc);
+
+    string output = get_named_writable_database_path("compact1-out");
+    rm_rf(output);
+
+    TEST_EXCEPTION(Xapian::InvalidOperationError,
+	db.compact(output));
+
+    db.commit();
+    db.compact(output);
+    db.close();
+
+    TEST_EQUAL(Xapian::Database::check(output, 0, &tout), 0);
+
+    return true;
+}
